import torch
import torch.nn as nn
import torch.nn.functional as F

class DoubleConv(nn.Module):
    """Double Convolution Block: (convolution => [BN] => ReLU) * 2"""

    def __init__(self, in_channels, out_channels, mid_channels=None):
        """
        Initialize DoubleConv module.

        Args:
            in_channels (int): Number of input channels.
            out_channels (int): Number of output channels.
            mid_channels (int, optional): Number of mid channels. Defaults to None.
        """
        super().__init__()
        if not mid_channels:
            mid_channels = out_channels
        self.double_conv = nn.Sequential(
            nn.Conv2d(in_channels, mid_channels, kernel_size=3, padding=1),
            nn.BatchNorm2d(mid_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(mid_channels, out_channels, kernel_size=3, padding=1),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True)
        )

    def forward(self, x):
        """
        Forward pass of DoubleConv module.

        Args:
            x (torch.Tensor): Input tensor.

        Returns:
            torch.Tensor: Output tensor.
        """
        return self.double_conv(x)

class Conv(nn.Module):
    """Convolution Block: (convolution => [BN] => ReLU)"""

    def __init__(self, in_channels, out_channels):
        """
        Initialize Conv module.

        Args:
            in_channels (int): Number of input channels.
            out_channels (int): Number of output channels.
        """
        super().__init__()
        self.conv = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
        )

    def forward(self, x):
        """
        Forward pass of Conv module.

        Args:
            x (torch.Tensor): Input tensor.

        Returns:
            torch.Tensor: Output tensor.
        """
        return self.conv(x)

class Down(nn.Module):
    """Downscaling with maxpool then double conv"""

    def __init__(self, in_channels, out_channels):
        """
        Initialize Down module.

        Args:
            in_channels (int): Number of input channels.
            out_channels (int): Number of output channels.
        """
        super().__init__()
        self.maxpool_conv = nn.Sequential(
            nn.MaxPool2d(2),
            DoubleConv(in_channels, out_channels)
        )

    def forward(self, x):
        """
        Forward pass of Down module.

        Args:
            x (torch.Tensor): Input tensor.

        Returns:
            torch.Tensor: Output tensor.
        """
        return self.maxpool_conv(x)

class Up(nn.Module):
    """Upscaling then double conv"""

    def __init__(self, in_channels, out_channels, bilinear=True):
        """
        Initialize Up module.

        Args:
            in_channels (int): Number of input channels.
            out_channels (int): Number of output channels.
            bilinear (bool, optional): Whether to use bilinear interpolation. Defaults to True.
        """
        super().__init__()

        # if bilinear, use the normal convolutions to reduce the number of channels
        self.up = F.interpolate
        self.conv = DoubleConv(in_channels, out_channels, in_channels // 2)

    def forward(self, x1, x2):
        """
        Forward pass of Up module.

        Args:
            x1 (torch.Tensor): Input tensor 1.
            x2 (torch.Tensor): Input tensor 2.

        Returns:
            torch.Tensor: Output tensor.
        """
        x1 = self.up(x1, x2.shape[2:], mode='bilinear')
        x = torch.cat([x2, x1], dim=1)
        return self.conv(x)

class UpInit(nn.Module):
    """Upscaling then double conv"""

    def __init__(self, temp_channels, in_channels, out_channels, bilinear=True, hax=False):
        """
        Initialize UpInit module.

        Args:
            temp_channels (int): Number of temporary channels.
            in_channels (int): Number of input channels.
            out_channels (int): Number of output channels.
            bilinear (bool, optional): Whether to use bilinear interpolation. Defaults to True.
            hax (bool, optional): Whether to apply special conditions. Defaults to False.
        """
        super().__init__()

        # if bilinear, use the normal convolutions to reduce the number of channels
        self.up = F.interpolate
        self.conv1 = Conv(in_channels, temp_channels)
        if hax:
            self.conv2 = DoubleConv(in_channels+temp_channels, out_channels, in_channels // 2)
        else:
            self.conv2 = DoubleConv(in_channels, out_channels, in_channels // 2)

    def forward(self, x1, x2):
        """
        Forward pass of UpInit module.

        Args:
            x1 (torch.Tensor): Input tensor 1.
            x2 (torch.Tensor): Input tensor 2.

        Returns:
            torch.Tensor: Output tensor.
        """
        x1 = self.conv1(x1)
        x1 = self.up(x1, x2.shape[2:], mode='bilinear')
        x = torch.cat([x2, x1], dim=1)
        return self.conv2(x)


class OutConv(nn.Module):
    """Output Convolution Block.

    This block consists of a single convolutional layer without activation.

    Args:
        in_channels (int): Number of input channels.
        out_channels (int): Number of output channels.
    """

    def __init__(self, in_channels, out_channels):
        super(OutConv, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=1)

    def forward(self, x):
        """Forward pass through the output convolution block.

        Args:
            x (torch.Tensor): Input tensor.

        Returns:
            torch.Tensor: Output tensor.
        """
        return self.conv(x)




import cv2
from PIL import Image
import numpy as np, torchvision
import torch, os, json
from skimage.color import label2rgb
from PIL import ImageColor
from cxas.label_mapper import label_mapper, id2label_dict, colors, colors_alpha, category_colors, category_ids
from cxas.file_io import FileLoader



def get_img(
            path: str
           ) -> np.array:
    """
    Load image from path.

    Parameters
    ----------
        path: Image file path
        
    Returns
    ----------
        image: Image in numpy format

    """
    assert os.path.isfile(path)
    loader  = FileLoader('')
    return loader.load_file(path)['orig_data']

def get_label(
            path: str
            ) -> np.array:
    """
    Load label from path with .npy-suffix.

    Parameters
    ----------
        path: label file path
        
    Returns
    ----------
        mask: Label mask in numpy format

    """
    assert os.path.isfile(path)
    return np.load(path)

def visualize_label(label: np.array, 
                    img: np.array, 
                    label_to_visualize: np.array, 
                    concat:bool = False, 
                    axis:int = 1
                   ) -> Image:
    """
    visualize certain labels from mask

    Parameters
    ----------
        label:  Mask in shape [classes (159), width, height]
        img:    Image in shape [classes (159), width, height]
        concat: Whether to display image and visualization side by side
        axis:   axis at which image and visualization are shown side by side
        
    Returns
    ----------
        visualization: Label visualization as PIL Image
        
    """
    

    out_mask = np.zeros((img.shape[0],img.shape[1],3)).astype(np.uint8)

    for i in label_to_visualize:
        imgray = (label[i,:,:]*255).astype(np.uint8)
        ret, thresh = cv2.threshold(imgray, 127, 255, 0)
        x = cv2.findContours(thresh, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        for contour in x:
            if (contour is not None) and (len(contour) > 0) and  (len(contour[0])>2) and (type(contour) == type(())):
                cv2.fillPoly(
                             out_mask, 
                             contour, 
                             # [int(j*255) for j in colors[i]],
                            [colors_alpha[i][0]*255,colors_alpha[i][1]*255,colors_alpha[i][2]*255,colors_alpha[i][3]],
                            )
        
    out_contour = np.zeros((img.shape[0],img.shape[1],3)).astype(np.uint8)
    for i in label_to_visualize:
        imgray = (label[i,:,:]*255).astype(np.uint8)
        ret, thresh = cv2.threshold(imgray, 127, 255, 0)
        x = cv2.findContours(thresh, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        for contour in x:
            if (contour is not None) and (len(contour) > 0) and  (len(contour[0])>2)  and (type(contour) == type(())):
                cv2.drawContours(out_contour, contour, -1, [int(j*255) for j in colors[i]], 2)
    
    out = cv2.addWeighted(out_contour, 1, out_mask, 0.7, 0.0)
    out = cv2.addWeighted(img, 0.5, out, 0.7, 0.0)
    
    if concat:
        return Image.fromarray(np.concatenate([img, out],axis)).convert('RGB')
    else:
        return Image.fromarray(out).convert('RGB')

def visualize_mask(class_names: list, 
                   mask:np.array,
                   image:np.array,
                   img_size: int,
                   cat:bool=True,
                   axis:int=1,
                  ) -> Image:
    """
    Resize image and label to desired size and visualize certain labels

    Parameters
    ----------
        class_names: List of classes of interest
        mask:  Mask in shape [classes (159), width, height]
        image:    Image in shape [3, width, height]
        img_size: Desired image size
        cat: Whether to display image and visualization side by side
        axis:   axis at which image and visualization are shown side by side
        
    Returns
    ----------
        visualization: Label visualization as PIL Image in desired size
        
    """
    # Resize image and label to desired size
    img = torch.nn.functional.interpolate(
                torch.tensor(image).float().unsqueeze(0),
                img_size,
                mode='bilinear'
            ).byte().numpy()[0]
        
        # reshape to match cv2 shapes
    img = np.transpose(img, [1,2,0])
    
    label = torch.nn.functional.interpolate(
                torch.tensor(mask).float().unsqueeze(0),
                img_size,
                mode='nearest'
            ).bool().numpy()[0]

    
    if type(class_names) == list:
        pass
    else:
        class_names = [class_names]
    
    label_to_visualize = np.concatenate([np.array(label_mapper[n]).flatten() for n in class_names]).flatten()
    
    return visualize_label(label, img,  label_to_visualize, cat, axis)

def visualize_from_file(class_names: list, 
                         img_path: str, 
                         label_path: str, 
                         img_size: int, 
                         cat:bool = True, 
                         axis:int = 1, 
                         do_store:bool = False,
                         out_dir:str = '',
                        ) -> Image:
    """
    Load Image and label, resize image and label to desired size, and visualize certain labels

    Parameters
    ----------
        class_names: list of class names to visualize
        img_path: path of the image file to visualize
        label_path: path of the label file to visualize
        img_size: size at which the image should be visualized
        cat: show image and label side by side
        axis: axis at which image and label are shown side by side
        do_store: boolean indicating whether to store the visualization in the out_dir 
                            with the associated label path and class_name
        out_dir: path at which to store visualization
        
    Returns
    ----------
        visualization: Pillow Image with labels overlaying original image
        
    """

    # Load image and label files
    img = get_img(img_path)
    label = get_label(label_path)
    # visualize desired labels
    visualization = visualize_mask(class_names, label, img, img_size, cat, axis)
    if do_store:
        os.makedirs(os.path.dirname(out_dir), exist_ok=True)
        visualization.save(os.path.join(
            out_dir,
            '{}_{}.png'.format(os.path.basename(label_path).split('.')[0], '_'.join(class_names))))
    return visualization


# https://github.com/chrise96/image-to-coco-json-converter/tree/master

from PIL import Image
import numpy as np
from skimage import measure
import os
import json


def create_category_annotation(category_dict):
    """
    Create category annotations in COCO JSON format.

    Args:
        category_dict (dict): Dictionary containing category names and IDs.

    Returns:
        list: List of category annotations.
    """
    category_list = []

    for key, value in category_dict.items():
        category = {
            "supercategory": key,
            "id": value,
            "name": key
        }
        category_list.append(category)

    return category_list

def get_coco_json_format():
    """
    Get the standard COCO JSON format.

    Returns:
        dict: COCO JSON format skeleton.
    """
    # Standard COCO format
    coco_format = {
        "info": {},
        "licenses": [],
        "images": [{}],
        "categories": [{}],
        "annotations": [{}]
    }

    return coco_format



import SimpleITK as sitk
import pydicom
import os
from SimpleITK import ImageSeriesReader
from pydicom import Dataset
from pydicom_seg import MultiClassWriter
import numpy as np
from pydicom_seg.template import from_dcmqi_metainfo


def write_dicom_seg(
    metainfo: str,
    # actually a directory, not a full path, though you can adjust it to make it
    # work with paths
    dcm_file: str,
    mask: np.array,
    out_dir: str,
    id_label_dict: dict,
) -> Dataset:
    """
    Writes DICOM Segmentation objects.

    :param metainfo: Path to metainfo.json.
    :param dcm_file: Directory to the stored series (or singular file).
    :param mask: Segmentation mask.
    :param out_dir: Output directory for DICOM Segmentation objects.
    :param id_label_dict: Dictionary mapping label indices to identifiers.
    :return: Pydicom dataset containing the DICOM SEG.
    """
    # Generate DICOM SEG template from metadata information
    template: Dataset = from_dcmqi_metainfo(metainfo)
    
    # Initialize MultiClassWriter for DICOM SEG creation
    writer: MultiClassWriter = MultiClassWriter(
        template=template,
        inplane_cropping=False,
        skip_empty_slices=True,
        skip_missing_segment=False
    )
    
    # Read DICOM series
    reader: ImageSeriesReader = sitk.ImageSeriesReader()
    dcm_files: tuple = (dcm_file,)
    reader.SetFileNames(dcm_files)
    image = reader.Execute()
    
    # Get shape of image data from SimpleITK image
    image_data_shape: tuple = sitk.GetArrayFromImage(image).shape
    
    # Ensure mask dimensions match image dimensions
    assert image_data_shape[1:] == mask.shape[1:], \
        f'Mask dimensions {mask.shape} and raw image dimensions {image_data_shape} do not match'
    
    # Iterate over slices in the mask
    for i in range(mask.shape[0]):
        # Skip if the mask slice is empty
        if mask[i:i+1].sum() == 0:
            continue
        
        # Extract segmentation data for the current slice
        segmentation_data = mask[i:i+1]
        segmentation = sitk.GetImageFromArray(segmentation_data)
        segmentation.CopyInformation(image)
        
        # Extract DICOM metadata from source images
        source_images = []
        for file in dcm_files:
            ds = pydicom.dcmread(file, stop_before_pixels=True)
            # Set default values if ImagePositionPatient or ImageOrientationPatient are missing
            try:
                _ = ds[0x0020, 0x0032]
            except KeyError:
                ds.ImagePositionPatient = [-212.025, -173.925, -0]
            
            try:
                _ = ds[0x0020, 0x0037]
            except KeyError:
                ds.ImageOrientationPatient = [1, -0, 0, -0, 1, -0]
            
            source_images.append(ds)
        
        # Write DICOM SEG and save to output directory
        out = writer.write(segmentation, source_images)
        out_path = os.path.join(out_dir, id_label_dict[str(i)] + '.dcm')
        out.save_as(out_path)



from sklearn.linear_model import LinearRegression
import numpy as np
from PIL import Image
from cxas.label_mapper import label_mapper
from cxas.extraction.func_helpers import get_perimeter_from_contour
from cxas.extraction.draw_helpers import draw_point, draw_line
import cv2
from cxas.label_mapper import id2label_dict


def get_all_perimeters(mask, img=None, draw = False):
    out = {}
    for i in range(mask.shape[0]):
        if mask[i].sum() == 0:
            perimeter = -1
        else:
            perimeter = get_indv_perimeter(mask[i].astype(np.uint8), 1, 1)
        out[id2label_dict[str(i)]+'_perimeter'] = perimeter
    return out 

def get_indv_perimeter(mask, spacing_x=1, spacing_y=1):
    contours, hierarchy = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
    # perimeter = cv2.arcLength(contours[0], True) * spacing_x
    perimeter = get_perimeter_from_contour(contours[0], spacing_x)
    if perimeter == 0:
        return np.array(-1)
    return perimeter



import torch
import torch.nn as nn
import torch.nn.functional as F
from .backbones import Backbone
from .unet_components import *

class BackboneUNet(nn.Module):
    def __init__(self, model_name: str, classes: int):
        """
        Initializes the BackboneUNet model.

        Args:
            model_name (str): Name of the backbone model.
            classes (int): Number of classes for segmentation.
        """
        super(BackboneUNet, self).__init__()
        
        # Initializing backbone and UNet head
        self.backbone = Backbone(model_name)
        self.head = get_unet_head(model_name, classes)
        self.dropout = nn.Dropout(p=0)
        self.threshold = 0.5

    def get_results(self, forward_dict, orig_dict):
        """
        Combines forward results with original input dictionary.

        Args:
            forward_dict (dict): Forward pass results.
            orig_dict (dict): Original input dictionary.

        Returns:
            dict: Combined dictionary with segmentation predictions.
        """
        out_dict = {**forward_dict, **orig_dict}

        out_dict['segmentation_preds'] = (forward_dict['logits'].sigmoid() > self.threshold).bool()
        
        return out_dict

    def forward(self, x):
        """
        Forward pass of the BackboneUNet model.

        Args:
            x (tensor): Input tensor.

        Returns:
            dict: Output dictionary with segmentation predictions.
        """
        img = x['data']
        forward_dict = self._forward(img)
        out_dict = self.get_results(forward_dict, x)
        return out_dict

    def _forward(self, x):
        """
        Internal forward pass of the BackboneUNet model.

        Args:
            x (tensor): Input tensor.

        Returns:
            dict: Dictionary containing features and logits.
        """
        backbone_dict = self.backbone(x)

        down1 = backbone_dict['feats_1_map']
        down2 = backbone_dict['feats_2_map']
        down3 = backbone_dict['feats_3_map']
        down4 = backbone_dict['feats_4_map']
        down5 = backbone_dict['feats_last_map']
        
        up1 = self.dropout(self.head.up1(down5, down4))
        up2 = self.dropout(self.head.up2(up1, down3))
        up3 = self.dropout(self.head.up3(up2, down2))
        up4 = self.dropout(self.head.up4(up3, down1))
        
        up = F.interpolate(up4, x.shape[2:], mode='bilinear')
        logits = self.head.out(up)
        
        return {'feats': up4, 'logits': logits}

def get_unet_head(network_name, classes, batch_size=1):
    """
    Retrieves the UNet head based on the network name.

    Args:
        network_name (str): Name of the network.
        classes (int): Number of classes.
        batch_size (int, optional): Batch size. Defaults to 1.

    Returns:
        nn.Module: UNet head module.
    """
    if 'vit' in network_name:
        return UNetHead(256, 128, classes, norm='batch' if batch_size > 1 else 'instance', constant=True)
    elif 'resnet34' in network_name:
        return UNetHead(512, 128, classes, norm='batch' if batch_size > 1 else 'instance')
    else:
        return UNetHead(2048, 128, classes, norm='batch' if batch_size > 1 else 'instance')

class UNetHead(nn.Module):
    def __init__(self, in_channels, ngf, num_classes, norm='batch', constant=False):
        """
        Initializes the UNet head.

        Args:
            in_channels (int): Number of input channels.
            ngf (int): Number of filters in the last layer.
            num_classes (int): Number of classes.
            norm (str, optional): Normalization type. Defaults to 'batch'.
            constant (bool, optional): Whether to use constant input size. Defaults to False.
        """
        super(UNetHead, self).__init__()

        if constant:
            self.up1 = UpInit(in_channels, in_channels, in_channels, True, True)
            self.up2 = Up(2 * in_channels, in_channels, True)
            self.up3 = Up(2 * in_channels, in_channels, True)
            self.up4 = Up(2 * in_channels, in_channels, True)

            self.out = OutConv(in_channels, num_classes)

        else:
            self.up1 = UpInit(in_channels // 2, in_channels, in_channels // 4, True)
            self.up2 = Up(in_channels // 4 * 2, in_channels // 8, True)
            self.up3 = Up(in_channels // 8 * 2, in_channels // 32, True)
            self.up4 = Up(in_channels // 32 * 2, ngf, True)

            self.out = OutConv(ngf, num_classes)



import unittest
from unittest import mock
import logging
import sys
import os
from pathlib import Path

# Adjust the import path based on your project structure
sys.path.insert(0, str(Path(__file__).resolve().parent.parent))  # Add the parent directory to sys.path

class TestCLI(unittest.TestCase):

    @mock.patch('argparse.ArgumentParser.parse_args')
    def test_parse_arguments(self, mock_parse_args):
        """Test argument parsing."""
        mock_parse_args.return_value = mock.Mock(
            input='/input/dummy_image.jpg',
            output='./output',
            output_type='png',
            gpus='0',
            model='UNet_ResNet50_default'
        )
        args = parse_arguments()
        self.assertEqual(args.input, '/input/dummy_image.jpg')
        self.assertEqual(args.output, './output')
        self.assertEqual(args.output_type, 'png')
        self.assertEqual(args.gpus, '0')
        self.assertEqual(args.model, 'UNet_ResNet50_default')

    @mock.patch('bin.cxas_segment.CXAS')  # Adjust to the actual import based on structure
    def test_process_file(self, mock_cxas):
        """Test processing a file."""
        mock_instance = mock_cxas.return_value
        mock_instance.process_file = mock.Mock()

        # Mocking sys.argv to simulate command line arguments
        sys.argv = ['cxas_segment', '-i', 'file.jpg', '-o', './output']

        main()  # Call the main function

        # Verify that process_file is called with the right arguments
        mock_instance.process_file.assert_called_once_with(
            filename='file.jpg',
            output_directory='./output',
            create=True,
            do_store=True,
            storage_type='png'  # Default output type
        )

    @mock.patch('bin.cxas_segment.CXAS')
    def test_process_directory(self, mock_cxas):
        """Test processing a directory."""
        mock_instance = mock_cxas.return_value
        mock_instance.process_folder = mock.Mock()

        # Mocking sys.argv to simulate command line arguments
        sys.argv = ['cxas_segment', '-i', './input', '-o', './output']

        main()  # Call the main function

        # Verify that process_folder is called with the right arguments
        mock_instance.process_folder.assert_called_once_with(
            input_directory_name='./input',
            output_directory='./output',
            create=True,
            storage_type='png'  # Default output type
        )

    @mock.patch('bin.cxas_segment.logging')
    def test_invalid_input_path(self, mock_logging):
        """Test error handling for invalid input path."""
        with mock.patch('pathlib.Path.is_file', return_value=False), \
             mock.patch('pathlib.Path.is_dir', return_value=False), \
             self.assertRaises(SystemExit):  # Catching SystemExit raised by the parser
             
            # Mocking sys.argv to simulate command line arguments
            sys.argv = ['cxas_segment', '-i', 'invalid_path', '-o', './output']
            main()

        # Check if the logging error method was called
        mock_logging.error.assert_called_with('invalid_path is neither a file nor a directory.')

if __name__ == '__main__':
    unittest.main()



import gdown, os, torch
from ..label_mapper import id2label_dict

model_urls = {
    'UNet_ResNet50_default': 'https://drive.google.com/file/d/1Y9zubvMzkYHoAqz-NvV6vniH5FKAF2iV/view?usp=drive_link'
}

def get_model(model_name, gpus=''):
    """
    Function to load a model by name and optionally move it to specified GPUs.

    Args:
        model_name (str): Name of the model to load.
        gpus (str): String containing GPU device IDs separated by commas.
                    If empty or 'cpu', the model will be loaded on CPU.

    Returns:
        torch.nn.Module: Loaded model.
    """
    assert model_name.split('_')[0] in list(model_getter.keys())

    model = model_getter[model_name.split('_')[0]](model_name)
    download_weights(model_name)
    model = load_weights(model, model_name, map_location='cpu' if 'cpu' in gpus else 'cuda')

    if 'cpu' not in gpus:
        gpus = [int(i) for i in gpus.split(',') if len(i) > 0]

        if len(gpus) > 1:
            assert torch.cuda.is_available()
            model.to(gpus[0])
            model = torch.nn.DataParallel(model, device_ids=gpus).cuda()
        elif len(gpus) == 1:
            assert torch.cuda.is_available()
            model.to(gpus[0])

    return model

def get_unet(model_name):
    """
    Function to get a U-Net model.

    Args:
        model_name (str): Name of the U-Net model.

    Returns:
        torch.nn.Module: U-Net model.
    """
    from .UNet.backbone_unet import BackboneUNet
    return BackboneUNet(model_name, len(id2label_dict.keys()))

def download_weights(model_name: str) -> None:
    """
    Function to download model weights.

    Args:
        model_name (str): Name of the model.
    """
    if "CXAS_PATH" in os.environ:
        store_path = os.path.join(os.environ['CXAS_PATH'], '.cxas')
    else:
        store_path = os.path.join(os.environ['HOME'], '.cxas')
    os.makedirs(os.path.join(store_path, 'weights/'), exist_ok=True)
    out_path = os.path.join(store_path, 'weights/{}'.format(model_name + '.pth'))
    if os.path.isfile(out_path):
        return
    else:
        gdown.download(model_urls[model_name], out_path, quiet=False, fuzzy=True)
        return

def load_weights(model, model_name: str, map_location: str = 'cuda:0'):
    """
    Function to load model weights.

    Args:
        model (torch.nn.Module): Model to load weights into.
        model_name (str): Name of the model.
        map_location (str): Location to map tensors to (default: 'cuda:0' if available, else 'cpu').

    Returns:
        torch.nn.Module: Model with loaded weights.
    """
    if "CXAS_PATH" in os.environ:
        store_path = os.path.join(os.environ['CXAS_PATH'], '.cxas')
    else:
        store_path = os.path.join(os.environ['HOME'], '.cxas')
    out_path = os.path.join(store_path, 'weights/{}'.format(model_name + '.pth'))
    assert os.path.isfile(out_path)

    checkpoint = torch.load(out_path, map_location=map_location)

    if 'module' in list(checkpoint['model'].keys())[0]:
        for i in list(checkpoint['model'].keys()):
            checkpoint['model'][i[len('module.'):]] = checkpoint['model'].pop(i)
    model.load_state_dict(checkpoint['model'], strict=False)
    return model

# Dictionary containing model getter functions
model_getter = {
    'UNet': get_unet,
}



from PIL import Image, ImageDraw

def draw_point(image, point, color, size):
    draw = ImageDraw.Draw(image)
    draw.ellipse([(point[0] - size / 2, point[1] - size / 2),
                  (point[0] + size / 2, point[1] + size / 2)],
                 fill=color)
    return image

def draw_line(image, p1, p2, color, width):
    draw = ImageDraw.Draw(image)
    # import pdb;pdb.set_trace()
    draw.line((p1, p2), fill=color, width=width)
    return image


from sklearn.linear_model import LinearRegression
import numpy as np
from PIL import Image
from cxas.label_mapper import label_mapper
from cxas.extraction.func_helpers import get_centers, sort_by_distance, get_min_dist
from cxas.extraction.draw_helpers import draw_point, draw_line

def get_spine_center_distance(mask, img=None, draw = False):
    """
    Calculate Spine-Center Distance. Distance from individual vertebrae to a regressed center line from all vertebrae.
    
    Parameters
    ----------
        mask: mask in form of np array [n_classes, width, height]
        img: source image, only used for visualization
        draw: whether to visualize the features
        
    Returns
    -------
        SCD: Distance from individual vertebrae to a regressed center line from all vertebrae.
    """
    def get_reg_line(points):
        x = np.array([c[0] for c in points])
        y = np.array([c[1] for c in points]).reshape(-1,1)

        reg = LinearRegression().fit( y, x)

        x_new = np.array(range(512)).reshape(-1,1)
        y_new = np.array([int(r) for r in reg.predict(x_new)])

        cc = np.array(list(zip(y_new,x_new[:,0])))

        return cc
    
    centers = get_centers(mask, label_mapper['all vertebrae'])
    centers = [c for c in centers if (c[0]>0) and (c[1] >0)]
    centers = sort_by_distance((256, 0), centers)

    cc = get_reg_line(centers)
    cc_ = [(c[0],c[1]) for c in cc]
    
    points, center_dists = get_min_dist(centers,cc)

    points = np.array(points)
    cc_ = [c for c in cc_ if (c[1]>centers[0][1]) and (c[1]<centers[-1][1])]
    
    
    # img, cc, 
    if draw:
        width = 8

        if img is None:
            img = Image.new("RGB", (mask.shape[1], mask.shape[2]), "black")
        else:
            pass

        for idx in range(len(cc_[:-1])):
            img = draw_line(img, cc_[idx], cc_[idx+1], '#2A9D8F',width)

        for idx in range(len(centers[:-1])):
            img = draw_line(img, centers[idx], centers[idx+1], '#F4A261',width)

        for idx in range(len(centers)):
            img = draw_line(img, centers[idx], (points[idx][0],points[idx][1]), '#264653',width)

        for idx in range(len(centers)):
            img = draw_point(img, centers[idx], '#E76F51',width*2)

        img = draw_point(img, cc_[0], '#264653',width*2)
        img = draw_point(img, cc_[-1], '#264653',width*2)
        
        return {'score':np.array(center_dists).mean(), 'drawing':img}
    return {'score':np.array(center_dists).mean()}